from __future__ import annotations

import os
import uuid
import time
from datetime import datetime, timezone
from typing import Dict, List, Optional, Union
from threading import Thread
from concurrent.futures import Future, ThreadPoolExecutor

from fastapi_poe import QueryRequest, PartialResponse

from weavel._constants import BACKEND_SERVER_URL
from weavel._api_client import APIClient
from weavel.utils import logger
from weavel.poe._poe_buffer_storage import PoeBufferStorage, PoeRequest, PoeSaveTraceDataBody

class PoeWorker:
    def __init__(
        self,
        api_key: str,
    ) -> None:
        self.api_key = api_key
        self.endpoint = BACKEND_SERVER_URL
        
        self.max_retry = 3
        self.flush_interval = 10
        self.flush_batch_size = 100
        
        self.api_client = APIClient()
        
        self.api_pool = ThreadPoolExecutor(max_workers=1)
        self.buffer_storage = PoeBufferStorage(max_buffer_size=1000)
        self._running = True
        self._thread = Thread(target=self.consume_buffer, daemon=True)
        self._thread.start()
    
    def log(
        self,
        user_request: QueryRequest,
        bot_responses: List[PartialResponse],
        response_timestamp: Optional[datetime] = datetime.now().isoformat(),
    ):
        pass
        user_log = PoeRequest(
            body=PoeSaveTraceDataBody(
                conversation_id=user_request.conversation_id,
                user_id=user_request.user_id,
                message_id=user_request.message_id,
                role="user",
                data_type="user_message",
                data_content=user_request.query[-1].content,
                timestamp=datetime.fromtimestamp(user_request.query[-1].timestamp / 1e6, timezone.utc).isoformat(),
            ),
        )
        bot_response_list: List[str] = [response.text for response in bot_responses]
        bot_response: str = "".join(bot_response_list)
        bot_log = PoeRequest(
            body=PoeSaveTraceDataBody(
                conversation_id=user_request.conversation_id,
                user_id=user_request.user_id,
                role="assistant",
                data_type="assistant_message",
                data_content=bot_response,
                timestamp=response_timestamp,
            ),
        )
        self.buffer_storage.push(user_log)
        self.buffer_storage.push(bot_log)
            
    def send_requests(self, requests: List[PoeRequest]):
        logger.info(requests)
        for attempt in range(self.max_retry):
            try:
                response = self.api_client.execute(
                    self.api_key,
                    self.endpoint,
                    "/poe/trace/batch",
                    method="POST",
                    json={"requests": [request.model_dump() for request in requests]},
                    timeout=10,
                )
                if response.status_code == 200:
                    return
            except Exception as e:
                print(e)
                time.sleep(2**attempt)
                continue
        
    def consume_buffer(self) -> None:
        while self._running:
            try:
                with self.buffer_storage.buffer_lock:
                    if self.buffer_storage.buffer_size == 0:
                        self.buffer_storage.not_empty_cv.wait(self.flush_interval)
                        continue
                    print("AWAKE!")
                    out = self.buffer_storage.pull(self.flush_batch_size)
                    future = self.api_pool.submit(self.send_requests, out)
                    future.add_done_callback(self._handle_api_task_completion)
                    self.buffer_storage.not_empty_cv.wait(self.flush_interval)
            except Exception as e:
                print(e)

    def flush(self) -> None:
        try:
            with self.buffer_storage.buffer_lock:
                if self.buffer_storage.buffer_size == 0:
                    return
                out = self.buffer_storage.pull_all()
                # for request chunks size of self.flush_batch_size
                for i in range(0, len(out), self.flush_batch_size):
                    request = out[i:i+self.flush_batch_size]
                    self.send_requests(request)
        except Exception as e:
            print(e)
            
    def stop(self) -> None:
        self._running = False
        with self.buffer_storage.buffer_lock:
            self.buffer_storage.not_empty_cv.notify()
        self._thread.join()
        self.flush()
                                     
    def _handle_api_task_completion(self, future: Future) -> None:
        try:
            future.result()
        except Exception as e:
            print(e)()


from pydantic import BaseModel, validator
from typing import Any, Dict, List, Optional, Sequence, Tuple, Union, Literal
from enum import Enum
from uuid import UUID
from datetime import datetime, date

class WeavelObject(BaseModel):
    """Weavel Object. Extended Pydantic BaseModel with support for UUID and datetime."""
    def __init__(self, **data: Any):
        for key, value in data.items():
            if key in self.__annotations__ and (self.__annotations__[key] == str or self.__annotations__[key] == Optional[str]):
                if (
                    isinstance(value, UUID)
                    or isinstance(value, datetime)
                    or isinstance(value, date)
                ):
                    try:
                        data[key] = str(value)
                    except:
                        data[key] = value
        super().__init__(**data)
        
        
class DataType(str, Enum):
    system_message = "system_message"
    user_message = "user_message"
    assistant_message = "assistant_message"
    inner_step = "inner_step"
    # retrieved_content = "retrieved_content"

class BackgroundTaskType(str, Enum):
    start_trace = "start_trace"
    log_trace_data = "log_trace_data"
    track_user = "track_user"
    # save_metadata_trace = "save_metadata_trace"
    create_tags_trace_data = "create_tags_trace_data"
    
class StartTraceBody(WeavelObject):
    """Start Trace body."""
    type: Literal['StartTraceBody'] = 'StartTraceBody'
    user_uuid: str
    trace_uuid: str
    timestamp: Optional[str] = None
    metadata: Optional[Dict[str, str]] = None
    
class TrackUserBody(WeavelObject):
    """Save metadata body."""
    type: Literal['TrackUserBody'] = 'TrackUserBody'
    user_uuid: str
    event_name: str
    properties: Dict[str, str]
    timestamp: Optional[str] = None

class SaveTraceDataBody(WeavelObject):
    """Log body."""
    type: Literal['SaveTraceDataBody'] = 'SaveTraceDataBody'
    user_uuid: str
    trace_uuid: str
    data_type: str
    data_content: str
    timestamp: Optional[str] = None
    unit_name: Optional[str] = None
    metadata: Optional[Dict[str, str]] = None
    
# class SaveMetadataTraceBody(WeavelObject):
#     """Save metadata body."""
#     trace_uuid: str
#     metadata: Dict[str, str]

class WeavelRequest(WeavelObject):
    """Weavel Request."""
    task: str
    # body: Union[StartTraceBody, SaveTraceDataBody, SaveMetadataTraceBody, TrackUserBody]
    body: Union[StartTraceBody, SaveTraceDataBody, TrackUserBody]
    
    @validator('task', pre=True)
    def validate_task(cls, v):
        if v not in BackgroundTaskType.__members__:
            raise ValueError(f"Invalid task type: {v}. Must be one of {list(BackgroundTaskType.__members__.keys())}")
        return v
    
class BatchRequest(WeavelObject):
    requests: list[WeavelRequest]



from __future__ import annotations

from threading import Condition, Lock
from typing import Dict, List

from weavel.types import WeavelRequest


class BufferStorage:
    def __init__(self, max_buffer_size: int):
        self.max_buffer_size = max_buffer_size
        self.buffer: List[WeavelRequest] = []
        self.buffer_lock = Lock()
        self.not_full_cv = Condition(self.buffer_lock)
        self.not_empty_cv = Condition(self.buffer_lock)

    @property
    def buffer_size(self) -> int:
        return len(self.buffer)

    def push(self, event: Dict) -> None:
        with self.buffer_lock:
            while self.buffer_size >= self.max_buffer_size:
                self.not_empty_cv.notify()  # immediately wake up possibly sleeping consumer
                self.not_full_cv.wait()
            self.buffer.append(event)
            if self.buffer_size >= self.max_buffer_size:
                self.not_empty_cv.notify()

    def pull(self, batch_size: int) -> List[WeavelRequest]:
        while not self.buffer_size:
            self.not_empty_cv.wait()
        pull_size = min(batch_size, self.buffer_size)
        out = self.buffer[:pull_size]
        self.buffer = self.buffer[pull_size:]
        self.not_full_cv.notify()
        return out

    def pull_all(self) -> List[WeavelRequest]:
        out = self.buffer[:]
        self.buffer = []
        self.not_full_cv.notify()
        return out



from weavel import create_client, WeavelClient, Trace

client: WeavelClient = create_client()

user_uuid = client.create_user_uuid()
print(user_uuid)

client.track(user_uuid, "identifier", {"name": "Jon Doe", "email": "jondoe@email.com"})
trace: Trace = client.start_trace(user_uuid) # -> trace_uuid


trace.log_message("system", "you are a helpful assistant.", unit_name="testapp")

from openai import OpenAI
openai_client = OpenAI()

user_message = "what can you do for me?"
trace.log_message("user", user_message, unit_name="testapp")

client.track(user_uuid, "subscription", {"plan": "premium"})
client.track(user_uuid, "paid", {"amount": "1000"})

# res = openai_client.chat.completions.create(
#     model="gpt-3.5-turbo",
#     messages=[
#         {"role": "system", "content": "You are a helpful assistant"},
#         {"role": "user", "content": user_message}
#     ]
# )

res = "I can assist you with a variety of tasks. I can help answer questions, provide information, give suggestions, assist with research, set reminders, manage your schedule, make reservations, provide translations, and much more. Just let me know what you need help with!"

# print(res.choices[0].message.content)

trace.log_message("assistant", res, unit_name="test_assistant")

# second_trace: Trace = client.start_trace(user_uuid)
# second_trace.log_message("user", "hello!", unit_name="second_testapp")
# second_trace.log_message("assistant", "I can help you with a variety of tasks.", unit_name="second_testapp")
# second_trace.log_message("system", "you are a helpful assistant.")

client.close()


from __future__ import annotations

import os
import uuid
import time
from datetime import datetime
from typing import Dict, List, Optional, Union
from threading import Thread
from concurrent.futures import Future, ThreadPoolExecutor

from weavel.types import (
    DataType,
    BackgroundTaskType,
    WeavelRequest,
    StartTraceBody,
    SaveTraceDataBody,
    TrackUserBody
)
from weavel._constants import BACKEND_SERVER_URL
from weavel._buffer_storage import BufferStorage
from weavel._api_client import APIClient
from weavel.utils import logger

class Worker:
    _instance = None

    def __new__(cls, *args, **kwargs):
        if not cls._instance:
            cls._instance = super(Worker, cls).__new__(cls)
        return cls._instance
    
    def __init__(
        self,
        api_key: str,
    ) -> None:
        if not hasattr(self, 'is_initialized'):
            self.api_key = api_key
            self.endpoint = BACKEND_SERVER_URL
            
            self.max_retry = 3
            self.flush_interval = 60
            self.flush_batch_size = 20
            
            self.api_client = APIClient()
            
            self.api_pool = ThreadPoolExecutor(max_workers=1)
            self.buffer_storage = BufferStorage(max_buffer_size=1000)
            self._running = True
            self._thread = Thread(target=self.consume_buffer, daemon=True)
            self._thread.start()
            self.is_initialized = True
        
    def _start_trace(self, trace_uuid: str, user_uuid: str, timestamp: Optional[datetime] = None, metadata: Optional[Dict[str, str]] = None) -> str:
        """Start the new trace for user_uuid.
        
        Returns:
            The trace UUID.
        """
        # add task to buffer
        request = WeavelRequest(**{
            "task": BackgroundTaskType.start_trace.value,
            "body" : StartTraceBody(**{
                "timestamp": str(timestamp or datetime.now().isoformat()),
                "trace_uuid": trace_uuid,
                "user_uuid": user_uuid,
                "metadata": metadata,
            })
        })
        self.buffer_storage.push(request)
        
        return trace_uuid
    
    def _track_users(self, user_uuid: str, name: str, properties: Dict) -> None:
        """Save the user event"""
        request = WeavelRequest(**{
            "task": BackgroundTaskType.track_user.value,
            "body" : TrackUserBody(**{
                "timestamp": str(datetime.now().isoformat()),
                "user_uuid": user_uuid,
                "event_name": name,
                "properties": properties,
            })
        })
        self.buffer_storage.push(request)
        
        return
        
    def user_message(
        self,
        user_uuid: str,
        trace_uuid: str,
        data_content: str,
        unit_name: Optional[str] = None,
        timestamp: Optional[datetime] = None,
        metadata: Optional[Dict[str, str]] = None, 
    ):
        """Log the "user_message" type data to the trace.
        
        Args:
            trace_uuid: The trace UUID.
            data_content: The data content.
            unit_name: The unit name.
            timestamp: The timestamp.
            metadata: The metadata.
        """
        request = WeavelRequest(**{
            "task": BackgroundTaskType.log_trace_data.value,
            "body" : SaveTraceDataBody(**{
                "timestamp": str(timestamp or datetime.now().isoformat()),
                "user_uuid": user_uuid,
                "trace_uuid": trace_uuid,
                "data_type": DataType.user_message,
                "data_content": data_content,
                "unit_name": unit_name,
                "metadata": metadata,
            })
        })
        self.buffer_storage.push(request)
        return
    
    def assistant_message(
        self,
        user_uuid: str,
        trace_uuid: str,
        data_content: str,
        unit_name: Optional[str] = None,
        timestamp: Optional[datetime] = None,
        metadata: Optional[Dict[str, str]] = None,
    ):
        """Log the "llm_background_message" type data to the trace.
        
        Args:
            trace_uuid: The trace UUID.
            data_content: The data content.
            unit_name: The unit name.
            timestamp: The timestamp.
            metadata: The metadata.
        """
        request = WeavelRequest(**{
            "task": BackgroundTaskType.log_trace_data.value,
            "body" : SaveTraceDataBody(**{
                "timestamp": str(timestamp or datetime.now().isoformat()),
                "user_uuid": user_uuid,
                "trace_uuid": trace_uuid,
                "data_type": DataType.assistant_message,
                "data_content": data_content,
                "unit_name": unit_name,
                "metadata": metadata,
            })
        })
        self.buffer_storage.push(request)
        return
    
    def system_message(
        self,
        user_uuid: str,
        trace_uuid: str,
        data_content: str,
        unit_name: Optional[str] = None,
        timestamp: Optional[datetime] = None,
        metadata: Optional[Dict[str, str]] = None,
    ):
        """Log the "system_message" type data to the trace.
        
        Args:
            trace_uuid: The trace UUID.
            data_content: The data content.
            unit_name: The unit name.
            timestamp: The timestamp.
            metadata: The metadata.
        """
        request = WeavelRequest(**{
            "task": BackgroundTaskType.log_trace_data.value,
            "body" : SaveTraceDataBody(**{
                "timestamp": str(timestamp or datetime.now().isoformat()),
                "user_uuid": user_uuid,
                "trace_uuid": trace_uuid,
                "data_type": DataType.system_message,
                "data_content": data_content,
                "unit_name": unit_name,
                "metadata": metadata,
            })
        })
        self.buffer_storage.push(request)
        return
    
    def inner_step(
        self,
        user_uuid: str,
        trace_uuid: str,
        data_content: str,
        unit_name: Optional[str] = None,
        timestamp: Optional[datetime] = None,
        metadata: Optional[Dict[str, str]] = None,
    ):
        """Log the "inner_step" type data to the trace.
        
        Args:
            trace_uuid: The trace UUID.
            data_content: The data content.
            unit_name: The unit name.
            timestamp: The timestamp.
            metadata: The metadata.
        """
        request = WeavelRequest(**{
            "task": BackgroundTaskType.log_trace_data.value,
            "body" : SaveTraceDataBody(**{
                "timestamp": str(timestamp or datetime.now().isoformat()),
                "user_uuid": user_uuid,
                "trace_uuid": trace_uuid,
                "data_type": DataType.inner_step,
                "data_content": data_content,
                "unit_name": unit_name,
                "metadata": metadata,
            })
        })
        self.buffer_storage.push(request)
        return
    
    def send_request(self, request: WeavelRequest):
        for attempt in range(self.max_retry):
            try:
                response = self.api_client.execute(
                    self.api_key,
                    self.endpoint,
                    "/trace",
                    method="POST",
                    json=request.model_dump(),
                    timeout=10,
                )
                if response.status_code == 200:
                    return
            except Exception as e:
                print(e)
                time.sleep(2**attempt)
                continue
            
    def send_requests(self, requests: List[WeavelRequest]):
        logger.info(requests)
        for attempt in range(self.max_retry):
            try:
                response = self.api_client.execute(
                    self.api_key,
                    self.endpoint,
                    "/trace/batch",
                    method="POST",
                    json={"requests": [request.model_dump() for request in requests]},
                    timeout=10,
                )
                if response.status_code == 200:
                    return
            except Exception as e:
                print(e)
                time.sleep(2**attempt)
                continue
        
    def consume_buffer(self) -> None:
        while self._running:
            try:
                with self.buffer_storage.buffer_lock:
                    if self.buffer_storage.buffer_size == 0:
                        self.buffer_storage.not_empty_cv.wait(self.flush_interval)
                        continue
                    out = self.buffer_storage.pull(self.flush_batch_size)
                    future = self.api_pool.submit(self.send_requests, out)
                    future.add_done_callback(self._handle_api_task_completion)
                    self.buffer_storage.not_empty_cv.wait(self.flush_interval)
            except Exception as e:
                print(e)

    def flush(self) -> None:
        try:
            with self.buffer_storage.buffer_lock:
                if self.buffer_storage.buffer_size == 0:
                    return
                out = self.buffer_storage.pull_all()
                # for request chunks size of self.flush_batch_size
                for i in range(0, len(out), self.flush_batch_size):
                    request = out[i:i+self.flush_batch_size]
                    self.send_requests(request)
        except Exception as e:
            print(e)
            
    def stop(self) -> None:
        self._running = False
        with self.buffer_storage.buffer_lock:
            self.buffer_storage.not_empty_cv.notify()
        self._thread.join()
        self.flush()
                                     
    def _handle_api_task_completion(self, future: Future) -> None:
        try:
            future.result()
        except Exception as e:
            print(e)


from __future__ import annotations

import os
import uuid
from datetime import datetime
from typing import Dict, List, Optional, Union

import pendulum
from dotenv import load_dotenv

from weavel.types import DataType
from weavel._worker import Worker

load_dotenv()

class Trace:
    def __init__(
        self,
        worker: Worker,
        user_uuid: str,
        trace_uuid: str,
    ):
        self.worker = worker
        self.user_uuid = user_uuid
        self.trace_uuid = trace_uuid
    
    def log_message(
        self,
        type: str, 
        content: str,
        timestamp: Optional[datetime] = None,
        unit_name: Optional[str] = None,
        metadata: Optional[Dict[str, str]] = None
    ):
        if type == "user":
            self.worker.user_message(self.user_uuid, self.trace_uuid, content, unit_name, timestamp, metadata)
        elif type == "assistant":
            self.worker.assistant_message(self.user_uuid, self.trace_uuid, content, unit_name, timestamp, metadata)
        elif type == "system":
            self.worker.system_message(self.user_uuid, self.trace_uuid, content, unit_name, timestamp, metadata)
        else:
            raise ValueError("Invalid message type.")
    
    def log_inner_step(
        self,
        content: str,
        timestamp: Optional[datetime] = None,
        unit_name: Optional[str] = None,
        metadata: Optional[Dict[str, str]] = None
    ):
        self.worker.inner_step(self.user_uuid, self.trace_uuid, content, unit_name, timestamp, metadata)
            
class WeavelClient:
    def __init__(
        self,
        api_key: Optional[str] = None,
    ):
        self.api_key = api_key or os.getenv("WEAVEL_API_KEY")
        assert self.api_key is not None, "API key not provided."
        self._worker = Worker(self.api_key)

    def create_user_uuid(
        self,
    ) -> str:
        """Create a new user_uuid.
        
        Returns:
            The user UUID.
        """
        return str(uuid.uuid4())
    
    def start_trace(
        self,
        user_uuid: str,
        timestamp: Optional[datetime] = None,
        metadata: Optional[Dict[str, str]] = None,
    ) -> Trace:
        """Start the new trace for user_uuid.
        
        Args:
            user_uuid: The user's UUID.
        Returns:
            The trace UUID.
        """
        trace_uuid = str(uuid.uuid4())
        self._worker._start_trace(trace_uuid, user_uuid, timestamp, metadata)
        trace = Trace(self._worker, user_uuid, trace_uuid)
        return trace
    
    def resume_trace(
        self,
        user_uuid,
        trace_uuid,
    ) -> Trace:
        """Resume the trace for (user_uuid, trace_uuid).
        Method for make Trace object for existing trace.
        """
        
        return Trace(self._worker, user_uuid, trace_uuid)
    
    def track(
        self,
        user_uuid: str,
        event_name: str,
        properties: Dict
    ):
        self._worker._track_users(user_uuid, event_name, properties)
    
    def close(
        self
    ):
        """Close the client."""
        self._worker.stop()
    
    
def create_client(
    api_key: Optional[str] = None,
) -> WeavelClient:
    """Create a Weavel client.
    
    Args:
        api_key: The API key.
    Returns:
        The Weavel client.
    """
    return WeavelClient(api_key=api_key)


from __future__ import annotations

import os
import uuid
from datetime import datetime
from typing import Dict, List, Optional, Union

import pendulum
from dotenv import load_dotenv
from fastapi_poe import QueryRequest, PartialResponse

from weavel.types import DataType
from weavel.poe._poe_worker import PoeWorker

load_dotenv()

class WeavelPoeClient:
    def __init__(
        self,
        api_key: Optional[str] = None,
    ):
        self.api_key = api_key or os.getenv("WEAVEL_API_KEY")
        assert self.api_key is not None, "API key not provided."
        self.worker = PoeWorker(self.api_key)
        
    def log(
        self,
        user_request: QueryRequest,
        bot_responses: List[PartialResponse],
        response_timestamp: Optional[datetime] = datetime.now().isoformat(),
    ):
        self.worker.log(user_request, bot_responses, response_timestamp)
        return
    
    def close(
        self
    ):
        """Close the client."""
        self.worker.stop()
    
    
def create_poe_client(
    api_key: Optional[str] = None,
) -> WeavelPoeClient:
    """Create a Weavel client for Poe integration.
    
    Args:
        api_key: The API key.
    Returns:
        The Weavel client.
    """
    return WeavelPoeClient(api_key=api_key)